import math

import astropy.units as astropy_units
import numpy as np
import scipy.stats as stats
from scipy.special import erf, erfcinv

from astromodels.functions.function import Function1D, FunctionMeta

deg2rad = np.pi / 180.0
rad2deg = 180.0 / np.pi

# noinspection PyPep8Naming


class Gaussian(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A Gaussian function

    latex : $ K \frac{1}{\sigma \sqrt{2 \pi}}\exp{\frac{(x-\mu)^2}{2~(\sigma)^2}} $

    parameters :

        F :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Normal
                    distribution
            initial value : 1

        mu :

            desc : Central value
            initial value : 0.0

        sigma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = 1.0 / math.sqrt(2 * np.pi)

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma):

        norm = self.__norm_const / sigma

        return F * norm * np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sigma, 2.0)))

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        mu = self.mu.value
        sigma = self.sigma.value

        sqrt_two = 1.414213562

        if x < 1e-16 or (1 - x) < 1e-16:

            res = -1e32

        else:

            res = mu + sigma * sqrt_two * erfcinv(2 * (1 - x))

        return res


class Truncated_gaussian(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A  truncated Gaussian function defined on the interval between the lower_bound
        (a) and upper_bound (b)

    latex : $\begin{split}f(x;\mu,\sigma,a,b)=\frac{\frac{1}{\sigma} \phi\left(
            \frac{x-\mu}{\sigma} \right)}{\Phi\left( \frac{b-\mu}{\sigma} \right)
            - \Phi\left( \frac{a-\mu}{\sigma} \right)}\\\phi\left(z\right)=
            \frac{1}{\sqrt{2 \pi}}\exp\left(-\frac{1}{2}z^2\right)\\\Phi\left(z\right)
            =\frac{1}{2}\left(1+erf\left(\frac{z}{\sqrt(2)}\right)\right)\end{split}$

    parameters :

        F :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Normal
                    distribution
            initial value : 1

        mu :

            desc : Central value
            initial value : 0.0

        sigma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

        lower_bound :

            desc: lower bound of gaussian, setting to -np.inf results in half normal
                    distribution
            initial value : -1.

        upper_bound :

            desc: upper bound of gaussian  setting to np.inf results in half normal
                    distribution
            initial value : 1.


    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = 1.0 / (math.sqrt(2 * np.pi))

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # The lower_bound has the same dimensions as the x
        self.lower_bound.unit = x_unit

        # The upper_bound has the same dimensions as the x
        self.upper_bound.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma, lower_bound, upper_bound):

        # phi is in unitless, so we need to do this trick
        # to keep the units right

        norm = self.__norm_const / sigma

        phi = np.zeros(x.shape) * F * norm * 0.0
        idx = (x >= lower_bound) & (x <= upper_bound)

        sqrt_two = 1.414213562

        # precalculate the arguments to the CDF

        lower_arg = (lower_bound - mu) / sigma
        upper_arg = (upper_bound - mu) / sigma

        # the typical gaussian functions

        phi[idx] = (
            np.exp(-np.power(x[idx] - mu, 2.0) / (2 * np.power(sigma, 2.0))) * F * norm
        )

        # the denominator is a function of the CDF

        if isinstance(F, astropy_units.Quantity):
            # erf cannot accept units

            upper_arg = upper_arg.value
            lower_arg = lower_arg.value

        theta_lower = 0.5 + 0.5 * erf(lower_arg / sqrt_two)

        theta_upper = 0.5 + 0.5 * erf(upper_arg / sqrt_two)

        return phi / (theta_upper - theta_lower)

    def from_unit_cube(self, x):

        mu = self.mu.value
        sigma = self.sigma.value
        lower_bound = self.lower_bound.value
        upper_bound = self.upper_bound.value

        sqrt_two = 1.414213562

        # if x < 1e-16 or (1 - x) < 1e-16:
        #    res = -1e32

        # precalculate the arguments to the  CDF

        lower_arg = (lower_bound - mu) / sigma
        upper_arg = (upper_bound - mu) / sigma

        theta_lower = 0.5 + 0.5 * erf(lower_arg / sqrt_two)

        theta_upper = 0.5 + 0.5 * erf(upper_arg / sqrt_two)

        # now precalculate the argument to the Inv. CDF

        arg = theta_lower + x * (theta_upper - theta_lower)

        out = mu + sigma * sqrt_two * erfcinv(2 * (1 - arg))

        return np.clip(out, lower_bound, upper_bound)


class Cauchy(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        The Cauchy distribution

    latex : $K\frac{1}{\gamma\pi}\left[\frac{\gamma^2}{(x-x_0)^2+\gamma^2}\right]$

    parameters :

        K :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Cauchy
                    distribution
            initial value : 1

        x0 :

            desc : Central value
            initial value : 0.0

        gamma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = 1.0 / math.sqrt(2 * np.pi)

    def _setup(self):
        self._is_prior = True

    def _set_units(self, x_unit, y_unit):
        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.K.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.x0.unit = x_unit

        # sigma has the same dimensions as x
        self.gamma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, K, x0, gamma):
        norm = 1 / (gamma * np.pi)

        gamma2 = gamma * gamma

        return K * norm * gamma2 / ((x - x0) * (x - x0) + gamma2)

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        x0 = self.x0.value
        gamma = self.gamma.value

        res = np.tan(np.pi * x - np.pi / 2) * gamma + x0

        return res


class Cosine_Prior(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A function which is constant on the interval angular interval of cosine

    latex : $\cos(x)$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : -90
            min : -np.inf
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 90
            min : -np.inf
            max : np.inf


        value :

            desc : Value in the interval
            initial value : 1.0



    """

    def _setup(self):

        self._fixed_units = (
            astropy_units.dimensionless_unscaled,
            astropy_units.dimensionless_unscaled,
        )

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # this prior needs to use the fixed units and
        # they do not need to be converted as they have no
        # dimension

        x_unit = self._fixed_units[0]
        y_unit = self._fixed_units[1]

        # Lower and upper bound has the same unit as x

        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def has_fixed_units(self):

        return True

    def evaluate(self, x, lower_bound, upper_bound, value):
        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x <= upper_bound)

        norm = (
            np.sin(deg2rad * (upper_bound)) - np.sin(deg2rad * (lower_bound))
        ) * 57.29577795

        result[idx] = value * np.cos(deg2rad * (x[idx])) / norm

        return result

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        cosdec_min = np.cos(deg2rad * (90.0 + self.lower_bound.value))
        cosdec_max = np.cos(deg2rad * (90.0 + self.upper_bound.value))

        v = x * (cosdec_max - cosdec_min)
        v += cosdec_min

        v = np.clip(v, -1.0, 1.0)
        # Now this generates on [0,pi)
        dec = np.arccos(v)

        # convert to degrees
        dec = rad2deg * dec
        # now in range [-90,90.0)
        dec -= 90.0

        return dec


class Log_normal(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A log normal function

    latex : $K\frac{1}{x\sigma\sqrt{2\pi}}\exp{\frac{(\log x/piv-\mu/piv)^2}
            {2~(\sigma)^2}} $

    parameters :

        F :
            desc : Integral between 0and +inf. Fix this to 1 to obtain a log Normal
                    distribution
            initial value : 1

        mu :

            desc : Central value
            initial value : 0.0

        sigma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

        piv :
            desc : pivot. Leave this to 1 for a proper log normal distribution
            initial value : 1.0
            fix : yes

    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    # Place this here to avoid recomputing it all the time

    __norm_const = 1.0 / math.sqrt(2 * np.pi)

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # The pivot has the same units as x
        self.piv.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma, piv):

        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * F * 0

        # The log normal is not defined if x < 0. The "0 * x" part is to conserve the
        # units if x has them, because 0 * x will be a Quantity with the same units as x
        idx = x > 0 * x

        result[idx] = (
            F
            * self.__norm_const
            / (sigma / piv * x[idx] / piv)
            * np.exp(
                -np.power(np.log(x[idx] / piv) - mu / piv, 2.0)
                / (2 * np.power(sigma / piv, 2.0)),
            )
        )

        return result

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        mu = self.mu.value
        sigma = self.sigma.value

        sqrt_two = 1.414213562

        if x < 1e-16 or (1 - x) < 1e-16:

            res = -1e32

        else:

            res = mu + sigma * sqrt_two * erfcinv(2 * (1 - x))

        return np.exp(res)


class Uniform_prior(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A function which is constant on the interval lower_bound - upper_bound and 0
        outside the interval. The extremes of the interval are counted as part of the
        interval.

    latex : $ f(x)=\begin{cases}0 & x < \text{lower_bound} \\\text{value} &
            \text{lower_bound} \le x \le \text{upper_bound} \\ 0 & x >
            \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 0
            min : -np.inf
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 1
            min : -np.inf
            max : np.inf

        value :

            desc : Value in the interval
            initial value : 1.0

    tests :
        - { x : 0.5, function value: 1.0, tolerance: 1e-20}
        - { x : -0.5, function value: 0, tolerance: 1e-20}

    """

    def _setup(self):
        self._is_prior = True

    def _set_units(self, x_unit, y_unit):
        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def evaluate(self, x, lower_bound, upper_bound, value):
        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x <= upper_bound)
        result[idx] = value

        return result

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        lower_bound = self.lower_bound.value
        upper_bound = self.upper_bound.value

        low = lower_bound
        spread = float(upper_bound - lower_bound)

        par = x * spread + low

        return par


class Log_uniform_prior(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A function which is K/x on the interval lower_bound - upper_bound and 0 outside
        the interval. The extremes of the interval are NOT counted as part of the
        interval. Lower_bound must be >= 0.

    latex : $ f(x)=K~\begin{cases}0 & x \le \text{lower_bound} \\\frac{1}{x} &
            \text{lower_bound} < x < \text{upper_bound} \\ 0 & x \ge \text{upper_bound}
            \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 1e-20
            min : 1e-30
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 100
            min : 1e-30
            max : np.inf

        K :

            desc : Normalization
            initial value : 1
            fix : yes

    """

    def _setup(self):

        self._is_prior = True
        self._handle_units = False

    def _set_units(self, x_unit, y_unit):
        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit
        self.K.unit = y_unit * x_unit

    def evaluate(self, x, lower_bound, upper_bound, K):
        # This makes the prior proper because it is the integral between lower_bound and
        # upper_bound

        res = np.where((x > lower_bound) & (x < upper_bound), K / x, 0)

        if isinstance(x, astropy_units.Quantity):

            return res * self.y_unit

        else:

            return res

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        low = math.log10(self.lower_bound.value)
        up = math.log10(self.upper_bound.value)

        spread = up - low
        par = 10 ** (x * spread + low)

        return par


class Beta(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A beta distribution function

    latex : $ f(x, a, b)=\frac{\Gamma(a+b) x^{a-1}(1-x)^{b-1}}{\Gamma(a) \Gamma(b)}$

    parameters :

        a :

            desc : first shape parameter
            initial value : 0.5
            min: 0.

        b :

            desc : second shape parameter
            initial value : 0.5
            min: 0.

    """

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        self.a.unit = astropy_units.dimensionless_unscaled
        self.b.unit = astropy_units.dimensionless_unscaled

    # noinspection PyPep8Naming
    def evaluate(self, x, a, b):

        return stats.beta.pdf(x, a, b)

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        a = self.a.value
        b = self.b.value

        return stats.beta.ppf(x, a, b)


class Gamma(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        A gamma distribution function

    latex : $ f(x,\alpha, \beta)=\frac{\beta^\alpha x^{\alpha-1}
            e^{-\beta x}}{\Gamma(\alpha)}$

    parameters :

        alpha :

            desc : first shape parameter
            initial value : 0.5
            min: 0.

        beta :

            desc : second shape parameter
            initial value : 1.
            min: 0.

    """

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

    # noinspection PyPep8Naming
    def evaluate(self, x, alpha, beta):

        return stats.gamma.pdf(x, alpha, scale=1.0 / beta)

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        alpha = self.alpha.value
        beta = self.beta.value

        return stats.gamma.ppf(x, alpha, scale=1.0 / beta)


class Exponential(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        An exponential distribution function

    latex : $ f(x, \alpha) = \alpha\exp(-\alpha x)$

    parameters :

        alpha :

            desc : first shape parameter
            initial value : 1
            min: 0.

    """

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        self.alpha.unit = astropy_units.dimensionless_unscaled

    # noinspection PyPep8Naming
    def evaluate(self, x, alpha):

        return stats.expon.pdf(x, scale=1.0 / alpha)

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        return stats.expon.ppf(x, scale=1.0 / self.alpha.value)


class Powerlaw_Prior(Function1D, metaclass=FunctionMeta):
    r"""
    description :

        An power law distribution function between a-b

    latex : $ f(x, \alpha) = \alpha x^{\alpha-1)$

    parameters :

        alpha :

            desc : slope parameter
            initial value : 1
            min: 0.
        a :
            desc: lower bound of distribution
            initial value : 0.
            min: 0.
        b:
            desc: upper bound of distribution
            initial value: 1
            min: 0.

    """

    def _setup(self):

        self._is_prior = True

    def _set_units(self, x_unit, y_unit):

        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.a.unit = astropy_units.dimensionless_unscaled
        self.b.unit = astropy_units.dimensionless_unscaled

    # noinspection PyPep8Naming
    def evaluate(self, x, alpha, a, b):

        d = b - a

        return stats.powerlaw.pdf(x, alpha, loc=a, scale=d)

    def from_unit_cube(self, x):
        """Used by multinest.

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        d = self.b.value - self.a.value

        return stats.powerlaw.ppf(x, self.alpha.value, loc=self.a.value, scale=d)



__author__ = "giacomov"

import collections

import numpy

from astromodels.core.spectral_component import SpectralComponent
from astromodels.core.tree import Node
from astromodels.core.units import get_units
from astromodels.sources.source import Source, SourceType
from astromodels.utils.logging import setup_logger
from astromodels.utils.pretty_list import dict_to_list

log = setup_logger(__name__)


class ParticleSource(Source, Node):
    """A source of particles with a certain distribution. This source does not
    produce any electromagnetic signal, but it is useful if used in conjuction
    with spectral shapes which need a particle distribution as input, like for
    example a Synchrotron kernel.

    :param name: name for the source
    :param distribution_shape: a function describing the energy
        distribution of the particles
    :param components: a list of SpectralComponents instances
    :return:
    """

    def __init__(self, name, distribution_shape=None, components=None):

        Node.__init__(self, name)

        if components is None:

            if distribution_shape is None:

                log.error(
                    "You have to either provied a list of components, or a "
                    "distribution shape"
                )

                raise AssertionError()

            components = [SpectralComponent("main", distribution_shape)]

        Source.__init__(self, components, SourceType.PARTICLE_SOURCE)

        # Add a node called 'spectrum'

        spectrum_node = Node("spectrum")
        spectrum_node._add_children(list(self._components.values()))

        self._add_child(spectrum_node)

        type(self).__call__ = type(self).get_flux

        # Set the units
        # Now sets the units of the parameters for the energy domain

        current_units = get_units()

        # energy as x and particle flux as y
        x_unit = current_units.energy
        y_unit = 1 / current_units.energy

        # Now set the units of the components
        for component in list(self._components.values()):
            component.shape.set_units(x_unit, y_unit)

    def get_flux(self, energies):
        """Get the total flux of this particle source at the given energies
        (summed over the components)"""

        results = [
            component.shape(energies) for component in list(self.components.values())
        ]

        return numpy.sum(results, 0)

    def _repr__base(self, rich_output=False):
        """Representation of the object.

        :param rich_output: if True, generates HTML, otherwise text
        :return: the representation
        """

        # Make a dictionary which will then be transformed in a list

        repr_dict = collections.OrderedDict()

        key = "%s (particle source)" % self.name

        repr_dict[key] = collections.OrderedDict()
        repr_dict[key]["spectrum"] = collections.OrderedDict()

        for component_name, component in list(self.components.items()):

            repr_dict[key]["spectrum"][component_name] = component.to_dict(minimal=True)

        return dict_to_list(repr_dict, rich_output)

    @property
    def free_parameters(self):
        """Returns a dictionary of free parameters for this source. We use the
        parameter path as the key because it's guaranteed to be unique, unlike
        the parameter name.

        :return:
        """
        free_parameters = collections.OrderedDict()

        for component in self._components.values():

            for par in component.shape.parameters.values():

                if par.free:

                    free_parameters[par.path] = par

        return free_parameters

    @property
    def parameters(self):
        """Returns a dictionary of all parameters for this source. We use the
        parameter path as the key because it's guaranteed to be unique, unlike
        the parameter name.

        :return:
        """
        all_parameters = collections.OrderedDict()

        for component in self._components.values():

            for par in component.shape.parameters.values():

                all_parameters[par.path] = par

        return all_parameters



import math
import os

from astromodels import update_logging_level
from astromodels.core.model import Model
from astromodels.core.model_parser import load_model
from astromodels.core.polarization import LinearPolarization, StokesPolarization
from astromodels.functions import Constant, Powerlaw
from astromodels.sources.point_source import PointSource

update_logging_level("DEBUG")


def test_linear_polarization_parameters():
    degree = 50.0
    angle = 30.0
    ps = PointSource(
        "PS",
        0,
        0,
        spectral_shape=Powerlaw(),
        polarization=LinearPolarization(degree=degree, angle=angle),
    )
    m1 = Model(ps)
    m1.display()

    m1.save("__test.yml", overwrite=True)

    mp = load_model("__test.yml")
    assert math.isclose(
        mp.sources["PS"].spectrum.main.polarization.degree.value, degree, rel_tol=0.02
    )
    assert math.isclose(
        mp.sources["PS"].spectrum.main.polarization.angle.value, angle, rel_tol=0.02
    )
    mp.display()

    os.remove("__test.yml")


def test_linear_polarization_functions():
    degree = Constant()
    angle = Constant()
    degree.k = 50
    angle.k = 30
    ps = PointSource(
        "PS",
        0,
        0,
        spectral_shape=Powerlaw(),
        polarization=LinearPolarization(degree=degree, angle=angle),
    )
    m1 = Model(ps)
    m1.display()

    m1.save("__test.yml", overwrite=True)

    mp = load_model("__test.yml")
    assert math.isclose(
        mp.sources["PS"].spectrum.main.polarization.degree.Constant.k.value,
        degree.k.value,
        rel_tol=0.02,
    )
    assert math.isclose(
        mp.sources["PS"].spectrum.main.polarization.angle.Constant.k.value,
        angle.k.value,
        rel_tol=0.02,
    )
    mp.display()

    os.remove("__test.yml")


def test_Stokes_polarization_functions():
    u = Constant()
    q = Constant()
    u.k = 0.5
    q.k = 0.5

    ps = PointSource(
        "PS", 0, 0, spectral_shape=Powerlaw(), polarization=StokesPolarization(Q=q, U=u)
    )
    m1 = Model(ps)
    m1.display()

    m1.save("__test.yml", overwrite=True)

    mp = load_model("__test.yml")
    assert math.isclose(
        mp.sources["PS"].spectrum.main.polarization.Q.Constant.k.value,
        q.k.value,
        rel_tol=0.02,
    )
    assert math.isclose(
        mp.sources["PS"].spectrum.main.polarization.Q.Constant.k.value,
        u.k.value,
        rel_tol=0.02,
    )
    mp.display()

    os.remove("__test.yml")



import collections
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Type

from rich.tree import Tree

from astromodels.utils.logging import setup_logger

from .cpickle_compatibility_layer import cPickle

log = setup_logger(__name__)


# This is necessary for pickle to be able to reconstruct a NewNode class (or derivate)
# during unpickling
class NewNodeUnpickler(object):
    def __call__(self, cls):

        instance = cls.__new__(cls)

        return instance


@dataclass(repr=False, unsafe_hash=True)
class NodeBase:
    _name: str
    _parent: Optional[Type["NodeBase"]] = field(repr=False, default=None)
    _children: Dict[str, Type["NodeBase"]] = field(
        default_factory=dict, repr=False, compare=False
    )
    _path: Optional[str] = field(repr=False, default="")

    # The next 3 methods are *really* necessary for anything to work

    def __reduce__(self):

        state = {}
        state["parent"] = self._get_parent()
        state["path"] = self._path
        state["children"] = self._get_children()
        state["child_names"] = [child.name for child in state["children"]]
        state["name"] = self._name
        state["__dict__"] = self.__dict__

        return NewNodeUnpickler(), (self.__class__,), state

    def __setstate__(self, state) -> None:

        self._children = {}

        # Set the name of this node

        self._name = state["name"]

        # Set the parent

        self._parent = state["parent"]

        # set the path

        self._path = state["path"]

        # Set the children
        # do this manually to avoid recursion
        # issues for children not yet built

        for child, name in zip(state["children"], state["child_names"]):

            self._children[name] = child

        # Restore everything else

        for k in state["__dict__"]:

            self.__dict__[k] = state["__dict__"][k]

    # This is necessary for copy.deepcopy to work
    def __deepcopy__(self, memodict={}):

        return cPickle.loads(cPickle.dumps(self))

    def _add_child(self, child: Type["NodeBase"]) -> None:

        if not isinstance(child, NodeBase):

            log.error(f"{child} is not of type Node")

            raise TypeError()

        log.debug_node(f"adding child {child._name}")

        if child._name not in self._children:

            # add the child to the dict

            self._children[child._name] = child

            # set the parent

            child._set_parent(self)

            # now go through and make sure
            # all the children know about the
            # new parent recursively

            child._update_child_path()

        else:

            log.error(f"A child with name {child._name} already exists")

            raise AttributeError()

    def _add_children(self, children: List[Type["NodeBase"]]) -> None:

        for child in children:
            self._add_child(child)

    def _remove_child(self, name: str, delete: bool = True) -> Optional["NodeBase"]:
        """Return a child."""

        # this kills the child

        if delete:

            del self._children[name]

            # return none

        # we want to keep the child
        # but orphan it

        else:

            child = self._children.pop(name)

            # now orphan the child

            child._orphan()

            # we want to get the
            # orphan back

            return child

            # return self._children.pop(name)

    def _orphan(self) -> None:
        """This will disconnect the current node from its parent and inform all
        the children about the change."""

        # disconnect from parent

        self._parent = None

        # be nice to the kids and tell them

        self._update_child_path()

    def _set_parent(self, parent: Type["NodeBase"]) -> None:
        """Set the parent and update path."""

        self._parent = parent

        parent_path = self._parent._get_path()

        if parent_path == "__root__":

            self._path = f"{self._name}"

        else:

            self._path = f"{parent_path}.{self._name}"

        log.debug_node(f"path is now: {self._path}")

    def _get_child(self, name: str) -> "NodeBase":
        """Return a child object."""

        return self._children[name]

    def _has_child(self, name: str) -> bool:
        """Is this child (name) in the tree."""
        return name in self._children

    def _get_children(self) -> Tuple["NodeBase"]:
        """Return a tuple of children."""

        return tuple(self._children.values())

    def _get_child_from_path(self, path: str) -> "NodeBase":
        """Get a child from a string path."""
        nodes = path.split(".")
        _next = self
        for node in nodes:
            _next = _next._get_child(node)

        return _next

    def __getitem__(self, key) -> "NodeBase":
        return self._get_child_from_path(key)

    def _recursively_gather_node_type(self, node, node_type) -> Dict[str, "NodeBase"]:

        instances = collections.OrderedDict()

        for child in node._get_children():

            # log.debug(f"on child {child._name}")

            if isinstance(child, node_type):

                path = child._get_path()

                # log.debug(f"on child {path}")

                instances[path] = child

                for sub_child in child._get_children():

                    instances.update(
                        self._recursively_gather_node_type(sub_child, node_type)
                    )

            else:

                instances.update(self._recursively_gather_node_type(child, node_type))

        return instances

    def _get_parent(self) -> "NodeBase":
        return self._parent

    def _get_path(self) -> "str":
        """Returns the str path of this node."""
        if self._parent is not None:
            return self._path

        else:
            return self._name

    def _root(self, source_only: bool = False) -> "NodeBase":
        """Returns the root of the node, will stop at the source if source_only
        is set to True."""
        if self.is_root:
            return self

        else:

            current_node = self

            # recursively walk up the tree to
            # the root

            while True:

                parent = current_node._parent

                if source_only:

                    if parent.name == "__root__":

                        return current_node

                current_node = current_node._parent

                if current_node.is_root:

                    return current_node

    @property
    def path(self) -> str:
        return self._get_path()

    def _update_child_path(self) -> None:
        """Update the path of all children recursively. This is needed if the
        name is changed.

        :returns:
        """
        # recursively update the path

        for name, child in self._children.items():

            child._path = f"{child._parent._get_path()}.{child._name}"

            if not child.is_leaf:

                child._update_child_path()

    def _change_name(self, name: str, clear_parent: bool = False) -> None:
        """Change the name of this node.

        This will have to update the children about the change. if
        clear_parent is provided, then the parent is removed
        """
        self._name = name

        if (self._parent is not None) and (not clear_parent):

            self._set_parent(self._parent)

        # update all the children
        self._update_child_path()

    @property
    def is_root(self) -> bool:
        """Is this the root of the tree."""
        return self._parent is None

    @property
    def is_leaf(self) -> bool:
        """Is this a a leaf of the tree."""
        if len(self._children) == 0:
            return True

        else:

            return False

    @property
    def name(self) -> str:
        return self._name

    def __getattr__(self, name):

        if name in self._children:

            return self._children[name]

        else:

            # log.error(f"Accessing an element {name} of the node that does not exist")

            raise AttributeError(
                f"Accessing an element {name} of the node that does not exist"
            )

            # return super(NodeBase).__getattr__(name)

    def __setattr__(self, name, value):

        # We cannot change a node
        # but if the node has a value
        # attribute, we want to call that

        if "_children" in self.__dict__:
            if name in self._children:

                if "_internal_value" in self._children[name].__dict__:

                    if not self._children[name].is_leaf:

                        log.warning(
                            f"Trying to set the value of a linked parameter ({name})"
                            " directly has no effect "
                        )

                        return

                    else:
                        # ok, this is likely  parameter

                        self._children[name].value = value

                else:

                    # this is going to be a node which
                    # we are not allowed to erase

                    # log.error(f"Accessing an element {name} of the node that does not
                    # exist")

                    raise AttributeError(
                        f"Accessing an element {name} of the node that does not exist"
                    )
            else:
                return super().__setattr__(name, value)
        else:

            return super().__setattr__(name, value)

    def plot_tree(self) -> Tree:
        """This plots the tree to the screen."""

        try:

            out = self.to_dict_with_types()
            name = "model"

        except AttributeError:

            out = self.to_dict()
            name = self.name

        tree = Tree(
            name,
            guide_style="bold medium_orchid",
            style="bold medium_orchid",
            highlight=True,
        )

        _recurse_dict(out, tree)

        return tree


def _recurse_dict(d: Dict[str, Any], tree: Tree, branch_color: Optional[str] = None):

    for k, v in d.items():

        if isinstance(v, collections.OrderedDict):

            if branch_color is not None:

                color = branch_color

            else:

                color = "not bold not blink medium_spring_green"

            if "position" in k:

                k = f"🔭 {k}"
                color = "bold not blink medium_spring_green"

            if "(point source)" in k:

                k = k.replace("(point source)", "")

                color = "bold blink medium_orchid"

                k = f"✨ {k}"

            if "(extended source)" in k:

                k = k.replace("(extended source)", "")

                color = "bold blink medium_orchid"

                k = f"🌌 {k}"

            if "spectrum" in k:

                color = "bold not blink light_goldenrod1"

                k = f"🌈 {k}"

                branch_color = "not bold not blink light_goldenrod1"

            if "main" in k:

                branch_color = "not bold not blink turquoise2"

            branch = tree.add(k, guide_style="bold not blink grey74", style=color)

            _recurse_dict(v, branch, branch_color=branch_color)

        else:

            pass

    return



import hashlib

import astropy.units as u
import numpy as np
from astropy import wcs
from astropy.coordinates import ICRS, BaseCoordinateFrame, SkyCoord
from astropy.io import fits

from astromodels.functions.function import Function2D, FunctionMeta
from astromodels.utils.angular_distance import angular_distance
from astromodels.utils.logging import setup_logger
from astromodels.utils.vincenty import vincenty

log = setup_logger(__name__)


class Latitude_galactic_diffuse(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        A Gaussian distribution in Galactic latitude around the Galactic plane

    latex : $ K \exp{\left( \frac{-b^2}{2 \sigma_b^2} \right)} $

    parameters :

        K :

            desc : normalization
            initial value : 1

        sigma_b :

            desc : Sigma for
            initial value : 1

        l_min :

            desc : min Longtitude
            initial value : 10

        l_max :

            desc : max Longtitude
            initial value : 30

    """

    # This is optional, and it is only needed if we need more setup after the
    # constructor provided by the meta class

    def _setup(self):

        self._frame = ICRS()

    def set_frame(self, new_frame):
        """Set a new frame for the coordinates (the default is ICRS J2000)

        :param new_frame: a coordinate frame from astropy
        :return: (none)
        """
        assert isinstance(new_frame, BaseCoordinateFrame)

        self._frame = new_frame

    def _set_units(self, x_unit, y_unit, z_unit):

        self.K.unit = z_unit
        self.sigma_b.unit = x_unit
        self.l_min.unit = y_unit
        self.l_max.unit = y_unit

    def evaluate(self, x, y, K, sigma_b, l_min, l_max):

        # We assume x and y are R.A. and Dec
        _coord = SkyCoord(ra=x, dec=y, frame=self._frame, unit="deg")

        b = _coord.transform_to("galactic").b.value
        l = _coord.transform_to("galactic").l.value

        return (
            K
            * np.exp(-(b**2) / (2 * sigma_b**2))
            * np.logical_or(
                np.logical_and(l > l_min, l < l_max),
                np.logical_and(l_min > l_max, np.logical_or(l > l_min, l < l_max)),
            )
        )

    def get_boundaries(self):

        max_b = self.sigma_b.max_value
        l_min = self.l_min.value
        l_max = self.l_max.value

        _coord = SkyCoord(
            l=[l_min, l_min, l_max, l_max],
            b=[max_b * -2.0, max_b * 2.0, max_b * 2.0, max_b * -2.0],
            frame="galactic",
            unit="deg",
        )

        # no dealing with 0 360 overflow
        min_lat = min(_coord.transform_to("icrs").dec.value)
        max_lat = max(_coord.transform_to("icrs").dec.value)
        min_lon = min(_coord.transform_to("icrs").ra.value)
        max_lon = max(_coord.transform_to("icrs").ra.value)

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        dL = (
            self.l_max.value - self.l_min.value
            if self.l_max.value > self.l_min.value
            else 360 + self.l_max.value - self.l_max.value
        )

        # integral -inf to inf exp(-b**2 / 2*sigma_b**2 ) db = sqrt(2pi)*sigma_b
        # Note that K refers to the peak diffuse flux (at b = 0) per square degree.
        integral = np.sqrt(2 * np.pi) * self.sigma_b.value * self.K.value * dL

        if isinstance(z, u.Quantity):
            z = z.value
        return integral * np.power(180.0 / np.pi, -2) * np.ones_like(z)


class Gaussian_on_sphere(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        A bidimensional Gaussian function on a sphere (in spherical coordinates)

    latex : $$ f(\vec{x}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{2\pi
            \sqrt{\det{\Sigma}}} \, {\rm exp}\left( -\frac{1}{2} (\vec{x}-
            \vec{x}_0)^\intercal \cdot \Sigma^{-1}\cdot (\vec{x}-\vec{x}_0)\right) \\
            \vec{x}_0 = ({\rm RA}_0,{\rm Dec}_0)\\ \Lambda = \left( \begin{array}{cc}
            \sigma^2 & 0 \\ 0 & \sigma^2 (1-e^2) \end{array}\right) \\ U = \left(
            \begin{array}{cc} \cos \theta & -\sin \theta \\ \sin \theta & cos \theta
            \end{array}\right) \\\Sigma = U\Lambda U^\intercal $$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        sigma :

            desc : Standard deviation of the Gaussian distribution
            initial value : 10
            min : 0
            max : 20

    """

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.sigma.unit = x_unit

    def evaluate(self, x, y, lon0, lat0, sigma):

        lon, lat = x, y

        angsep = angular_distance(lon0, lat0, lon, lat)

        s2 = sigma**2

        return (
            (180 / np.pi) ** 2 * 1 / (2.0 * np.pi * s2) * np.exp(-0.5 * angsep**2 / s2)
        )

    def get_boundaries(self):

        # Truncate the gaussian at 2 times the max of sigma allowed

        max_sigma = self.sigma.max_value

        min_lat = max(-90.0, self.lat0.value - 2 * max_sigma)
        max_lat = min(90.0, self.lat0.value + 2 * max_sigma)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if (
            max_abs_lat > 89.0
            or 2 * max_sigma / np.cos(max_abs_lat * np.pi / 180.0) >= 180.0
        ):

            min_lon = 0.0
            max_lon = 360.0

        else:

            min_lon = self.lon0.value - 2 * max_sigma / np.cos(
                max_abs_lat * np.pi / 180.0
            )
            max_lon = self.lon0.value + 2 * max_sigma / np.cos(
                max_abs_lat * np.pi / 180.0
            )

            if min_lon < 0.0:

                min_lon = min_lon + 360.0

            elif max_lon > 360.0:

                max_lon = max_lon - 360.0

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Asymm_Gaussian_on_sphere(Function2D, metaclass=FunctionMeta):
    r"""
    description :
       A bidimensional Gaussian function on a sphere (in spherical coordinates) see
       https://en.wikipedia.org/wiki/Gaussian_function#Two-dimensional_Gaussian_function

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        a :

            desc : Standard deviation of the Gaussian distribution (major axis)
            initial value : 10
            min : 0
            max : 20

        e :

            desc : Excentricity of Gaussian ellipse
            initial value : 0.9
            min : 0
            max : 1

        theta :

            desc : inclination of major axis to a line of constant latitude
            initial value : 10.
            min : -90.0
            max : 90.0

    """

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and a have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.a.unit = x_unit
        self.e.unit = u.dimensionless_unscaled
        self.theta.unit = u.degree

    def evaluate(self, x, y, lon0, lat0, a, e, theta):

        lon, lat = x, y

        b = a * np.sqrt(1.0 - e**2)

        dX = np.atleast_1d(angular_distance(lon0, lat0, lon, lat0))
        dY = np.atleast_1d(angular_distance(lon0, lat0, lon0, lat))

        dlon = lon - lon0
        if isinstance(dlon, u.Quantity):
            dlon = (dlon.to(u.degree)).value

        idx = np.logical_and(
            np.logical_or(dlon < 0, dlon > 180),
            np.logical_or(dlon > -180, dlon < -360),
        )
        dX[idx] = -dX[idx]

        idx = lat < lat0
        dY[idx] = -dY[idx]

        if isinstance(theta, u.Quantity):
            phi = (theta.to(u.degree)).value + 90.0
        else:
            phi = theta + 90.0

        cos2_phi = np.power(np.cos(phi * np.pi / 180.0), 2)
        sin2_phi = np.power(np.sin(phi * np.pi / 180.0), 2)

        sin_2phi = np.sin(2.0 * phi * np.pi / 180.0)

        A = cos2_phi / (2.0 * b**2) + sin2_phi / (2.0 * a**2)

        B = -sin_2phi / (4.0 * b**2) + sin_2phi / (4.0 * a**2)

        C = sin2_phi / (2.0 * b**2) + cos2_phi / (2.0 * a**2)

        E = -A * np.power(dX, 2) + 2.0 * B * dX * dY - C * np.power(dY, 2)

        return np.power(180 / np.pi, 2) * 1.0 / (2 * np.pi * a * b) * np.exp(E)

    def get_boundaries(self):

        # Truncate the gaussian at 2 times the max of sigma allowed

        min_lat = max(-90.0, self.lat0.value - 2 * self.a.max_value)
        max_lat = min(90.0, self.lat0.value + 2 * self.a.max_value)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if (
            max_abs_lat > 89.0
            or 2 * self.a.max_value / np.cos(max_abs_lat * np.pi / 180.0) >= 180.0
        ):

            min_lon = 0.0
            max_lon = 360.0

        else:

            min_lon = self.lon0.value - 2 * self.a.max_value / np.cos(
                max_abs_lat * np.pi / 180.0
            )
            max_lon = self.lon0.value + 2 * self.a.max_value / np.cos(
                max_abs_lat * np.pi / 180.0
            )

            if min_lon < 0.0:

                min_lon = min_lon + 360.0

            elif max_lon > 360.0:

                max_lon = max_lon - 360.0

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Disk_on_sphere(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        A bidimensional disk/tophat function on a sphere (in spherical coordinates)

    latex : $$ f(\vec{x}) = \left(\frac{180}{\pi}\right)^2 \frac{1}{\pi~({\rm
            radius})^2} ~\left\{\begin{matrix} 1 & {\rm if}& {\rm | \vec{x} - \vec{x}_0|
            \le {\rm radius}} \\ 0 & {\rm if}& {\rm | \vec{x} - \vec{x}_0| > {\rm
            radius}} \end{matrix}\right. $$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        radius :

            desc : Radius of the disk
            initial value : 15
            min : 0
            max : 20

    """

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.radius.unit = x_unit

    def evaluate(self, x, y, lon0, lat0, radius):

        lon, lat = x, y

        angsep = angular_distance(lon0, lat0, lon, lat)

        return np.power(180 / np.pi, 2) * 1.0 / (np.pi * radius**2) * (angsep <= radius)

    def get_boundaries(self):

        # Truncate the disk at 2 times the max of radius allowed

        max_radius = self.radius.max_value

        min_lat = max(-90.0, self.lat0.value - 2 * max_radius)
        max_lat = min(90.0, self.lat0.value + 2 * max_radius)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if (
            max_abs_lat > 89.0
            or 2 * max_radius / np.cos(max_abs_lat * np.pi / 180.0) >= 180.0
        ):

            min_lon = 0.0
            max_lon = 360.0

        else:

            min_lon = self.lon0.value - 2 * max_radius / np.cos(
                max_abs_lat * np.pi / 180.0
            )
            max_lon = self.lon0.value + 2 * max_radius / np.cos(
                max_abs_lat * np.pi / 180.0
            )

            if min_lon < 0.0:

                min_lon = min_lon + 360.0

            elif max_lon > 360.0:

                max_lon = max_lon - 360.0

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Ellipse_on_sphere(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        An ellipse function on a sphere (in spherical coordinates)

    latex : $$ f(\vec{x}) = \left(\frac{180}{\pi}\right)^2 \frac{1}{\pi~ a b}
            ~\left\{\begin{matrix} 1 & {\rm if}& {\rm | \vec{x} - \vec{x}_{f1}|
            + | \vec{x} - \vec{x}_{f2}| \le {\rm 2a}} \\ 0 & {\rm if}& {\rm | \vec{x}
            - \vec{x}_{f1}| + | \vec{x}-\vec{x}_{f2}| > {\rm 2a}} \end{matrix}\right. $$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        a :

            desc : semimajor axis of the ellipse
            initial value : 15.
            min : 0
            max : 20

        e :

            desc : eccentricity of ellipse
            initial value : 0.9
            min : 0
            max : 1

        theta :

            desc : inclination of semimajoraxis to a line of constant latitude
            initial value : 0.0
            min : -90.0
            max : 90.0
    """

    lon1 = None
    lat1 = None
    lon2 = None
    lat2 = None
    focal_pts = False

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 have most probably all units of degrees.
        # However, let's set them up here just to save for the possibility of
        # using the formula with other units (although it is probably never
        # going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.a.unit = x_unit
        # eccentricity is dimensionless
        self.e.unit = u.dimensionless_unscaled
        self.theta.unit = u.degree

    def calc_focal_pts(self, lon0, lat0, a, b, theta):
        # focal distance
        f = np.sqrt(a**2 - b**2)

        if isinstance(theta, u.Quantity):
            bearing = 90.0 - (theta.to(u.degree)).value
        else:
            bearing = 90.0 - theta

        # coordinates of focal points
        lon1, lat1 = vincenty(lon0, lat0, bearing, f)
        lon2, lat2 = vincenty(lon0, lat0, bearing + 180.0, f)

        return lon1, lat1, lon2, lat2

    def evaluate(self, x, y, lon0, lat0, a, e, theta):

        b = a * np.sqrt(1.0 - e**2)

        # calculate focal points

        self.lon1, self.lat1, self.lon2, self.lat2 = self.calc_focal_pts(
            lon0, lat0, a, b, theta
        )
        self.focal_pts = True

        # lon/lat of point in question
        lon, lat = x, y

        # sum of geodesic distances to focii (should be <= 2a to be in ellipse)
        angsep1 = angular_distance(self.lon1, self.lat1, lon, lat)
        angsep2 = angular_distance(self.lon2, self.lat2, lon, lat)
        angsep = angsep1 + angsep2

        return np.power(180 / np.pi, 2) * 1.0 / (np.pi * a * b) * (angsep <= 2 * a)

    def get_boundaries(self):

        # Truncate the ellipse at 2 times the max of semimajor axis allowed

        max_radius = self.a.max_value

        min_lat = max(-90.0, self.lat0.value - 2 * max_radius)
        max_lat = min(90.0, self.lat0.value + 2 * max_radius)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if (
            max_abs_lat > 89.0
            or 2 * max_radius / np.cos(max_abs_lat * np.pi / 180.0) >= 180.0
        ):

            min_lon = 0.0
            max_lon = 360.0

        else:

            min_lon = self.lon0.value - 2 * max_radius / np.cos(
                max_abs_lat * np.pi / 180.0
            )
            max_lon = self.lon0.value + 2 * max_radius / np.cos(
                max_abs_lat * np.pi / 180.0
            )

            if min_lon < 0.0:

                min_lon = min_lon + 360.0

            elif max_lon > 360.0:

                max_lon = max_lon - 360.0

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class SpatialTemplate_2D(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        User input Spatial Template.  Expected to be normalized to 1/sr

    latex : $ hi $

    parameters :

        K :

            desc : normalization
            initial value : 1
            fix : yes
        hash :

            desc: hash of model map [needed for memoization]
            initial value: 1
            fix: yes
        ihdu:
            desc: header unit index of fits file
            initial value: 0
            fix: True
            min: 0

    properties:
        fits_file:
            desc: fits file to load
            defer: True
            function: _load_file
        frame:
            desc: coordinate frame
            initial value: icrs
            allowed values:
            - icrs
            - galactic
            - fk5
            - fk4
            - fk4_no_e
    """

    def _set_units(self, x_unit, y_unit, z_unit):

        self.K.unit = z_unit

    # This is optional, and it is only needed if we need more setup after the
    # constructor provided by the meta class

    # def _setup(self):

    # self._frame = "icrs"
    # self._fitsfile = None
    # self._map = None

    def _load_file(self):

        self._fitsfile = self.fits_file.value

        with fits.open(self._fitsfile) as f:

            self._wcs = wcs.WCS(header=f[int(self.ihdu.value)].header)
            self._map = f[int(self.ihdu.value)].data

            self._nX = f[int(self.ihdu.value)].header["NAXIS1"]
            self._nY = f[int(self.ihdu.value)].header["NAXIS2"]

            # note: map coordinates are switched compared to header. NAXIS1 is
            # coordinate 1, not 0. see
            # http://docs.astropy.org/en/stable/io/fits/#working-with-image-data
            assert (
                self._map.shape[1] == self._nX
            ), "NAXIS1 = %d in fits header, but %d in map" % (
                self._nX,
                self._map.shape[1],
            )
            assert (
                self._map.shape[0] == self._nY
            ), "NAXIS2 = %d in fits header, but %d in map" % (
                self._nY,
                self._map.shape[0],
            )

            # test if the map is normalized as expected
            area = wcs.utils.proj_plane_pixel_area(self._wcs)
            dOmega = (area * u.deg * u.deg).to(u.sr).value
            total = self._map.sum() * dOmega

            if not np.isclose(total, 1, rtol=1e-2):
                log.warning(
                    "2D template read from {} is normalized to {} (expected: 1)".format(
                        self._fitsfile, total
                    )
                )

            # hash sum uniquely identifying the template function (defined by its 2D map
            # array and coordinate system) this is needed so that the memoization won't
            # confuse different SpatialTemplate_2D objects.
            h = hashlib.sha224()
            h.update(self._map)
            h.update(repr(self._wcs).encode("utf-8"))
            self.hash = int(h.hexdigest(), 16)

    # def to_dict(self, minimal=False):

    #      data = super(Function2D, self).to_dict(minimal)

    #      if not minimal:

    #         data['extra_setup'] = {"_fitsfile": self._fitsfile,"_frame":self._frame}

    #      return data

    # def set_frame(self, new_frame):
    #     """
    #         Set a new frame for the coordinates (the default is ICRS J2000)

    #         :param new_frame: a coordinate frame from astropy
    #         :return: (none)
    #         """
    #     assert new_frame.lower() in ['icrs', 'galactic', 'fk5', 'fk4', 'fk4_no_e' ]

    #     self._frame = new_frame

    def evaluate(self, x, y, K, hash, ihdu):

        # X and Y are defined by the frame (ICRS,galactic, etc..)
        coord = SkyCoord(x, y, frame=self.frame.value, unit="deg")

        # transform input coordinates to pixel coordinates;
        # SkyCoord takes care of necessary coordinate frame transformations.
        Xpix, Ypix = coord.to_pixel(self._wcs)

        Xpix = np.atleast_1d(Xpix.astype(int))
        Ypix = np.atleast_1d(Ypix.astype(int))

        # find pixels that are in the template ROI, otherwise return zero
        # iz = np.where((Xpix<self._nX) & (Xpix>=0) & (Ypix<self._nY) & (Ypix>=0))[0]
        iz = (Xpix < self._nX) & (Xpix >= 0) & (Ypix < self._nY) & (Ypix >= 0)
        out = np.zeros_like(Xpix).astype(float)
        out[iz] = self._map[Ypix[iz], Xpix[iz]]

        return np.multiply(K, out)

    def get_boundaries(self):

        # if self._map is None:

        #     self.load_file(self._fitsfile)

        # We use the max/min RA/Dec of the image corners to define the boundaries.
        # Use the 'outside' of the pixel corners, i.e. from pixel 0 to nX in 0-indexed
        # accounting.

        Xcorners = np.array([0, 0, self._nX, self._nX])
        Ycorners = np.array([0, self._nY, 0, self._nY])

        corners = SkyCoord.from_pixel(
            Xcorners, Ycorners, wcs=self._wcs, origin=0
        ).transform_to(self.frame.value)

        min_lon = min(corners.ra.degree)
        max_lon = max(corners.ra.degree)

        min_lat = min(corners.dec.degree)
        max_lat = max(corners.dec.degree)

        return (min_lon, max_lon), (min_lat, max_lat)

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.multiply(self.K.value, np.ones_like(z))


class Power_law_on_sphere(Function2D, metaclass=FunctionMeta):
    r"""
    description :

        A power law function on a sphere (in spherical coordinates)

    latex : $$ f(\vec{x}) = \left(\frac{180}{\pi}\right)^{-1.*index}  \left\{
            \begin{matrix} 0.05^{index} & {\rm if} & ||\vec{x}-\vec{x}_0|| \le 0.05\\
            ||\vec{x}-\vec{x}_0||^{index} & {\rm if} & 0.05 < ||\vec{x}-\vec{x}_0|| \le
            maxr \\ 0 & {\rm if} & ||\vec{x}-\vec{x}_0||>maxr\end{matrix}\right. $$

    parameters :

        lon0 :

            desc : Longitude of the center of the source
            initial value : 0.0
            min : 0.0
            max : 360.0

        lat0 :

            desc : Latitude of the center of the source
            initial value : 0.0
            min : -90.0
            max : 90.0

        index :

            desc : power law index
            initial value : -2.0
            min : -5.0
            max : -1.0

        maxr :

            desc : max radius
            initial value : 20.
            fix : yes

        minr :

            desc : radius below which the PL is approximated as a constant
            initial value : 0.05
            fix : yes

    """

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.index.unit = u.dimensionless_unscaled
        self.maxr.unit = x_unit
        self.minr.unit = x_unit

    def evaluate(self, x, y, lon0, lat0, index, maxr, minr):

        lon, lat = x, y

        angsep = angular_distance(lon0, lat0, lon, lat)

        if maxr <= minr:
            norm = np.power(np.pi / 180.0, 2.0 + index) * np.pi * maxr**2 * minr**index
        elif self.index.value == -2.0:
            norm = np.pi * (1.0 + 2.0 * np.log(maxr / minr))
        else:
            norm = np.power(minr * np.pi / 180.0, 2.0 + index) * np.pi + 2.0 * np.pi / (
                2.0 + index
            ) * (
                np.power(maxr * np.pi / 180.0, index + 2.0)
                - np.power(minr * np.pi / 180.0, index + 2.0)
            )

        value = (
            np.less_equal(angsep, maxr)
            * np.power(np.pi / 180.0, index)
            * np.power(
                np.add(
                    np.multiply(angsep, np.greater(angsep, minr)),
                    np.multiply(minr, np.less_equal(angsep, minr)),
                ),
                index,
            )
        )

        return value / norm

    def get_boundaries(self):

        return (
            (self.lon0.value - self.maxr.value),
            (self.lon0.value + self.maxr.value),
        ), (
            (self.lat0.value - self.maxr.value),
            (self.lat0.value + self.maxr.value),
        )

    def get_total_spatial_integral(self, z=None):
        """Returns the total integral (for 2D functions) or the integral over
        the spatial components (for 3D functions). needs to be implemented in
        subclasses.

        :return: an array of values of the integral (same dimension as
            z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


# class FunctionIntegrator(Function2D):
#     r"""
#         description :
#
#             Returns the average of the integrand function (a 1-d function) over the
#             interval x-y. The integrand is set
#             using the .integrand property, like in:
#
#             > G = FunctionIntegrator()
#             > G.integrand = Powerlaw()
#
#         latex : $$ G(x,y) = \frac{\int_{x}^{y}~f(x)~dx}{y-x}$$
#
#         parameters :
#
#             s :
#
#                 desc : if s=0, then the integral will *not* be normalized by (y-x),
#                        otherwise (default) it will.
#                 initial value : 1
#                 fix : yes
#         """
#
#     __metaclass__ = FunctionMeta
#
#     def _set_units(self, x_unit, y_unit, z_unit):
#
#         # lon0 and lat0 and rdiff have most probably all units of degrees. However,
#         # let's set them up here just to save for the possibility of using the
#         # formula with other units (although it is probably never going to happen)
#
#         self.s = u.dimensionless_unscaled
#
#     def evaluate(self, x, y, s):
#
#         assert y-x >= 0, "Cannot obtain the integral if the integration interval is
#                           zero or negative!"
#
#         integral = self._integrand.integral(x, y)
#
#         if s==0:
#
#             return integral
#
#         else:
#
#             return integral / (y-x)
#
#
#     def get_boundaries(self):
#
#         return (-np.inf, +np.inf), (-np.inf, +np.inf)
#
#     def _set_integrand(self, function):
#
#         self._integrand = function
#
#     def _get_integrand(self):
#
#         return self._integrand
#
#     integrand = property(_get_integrand, _set_integrand,
#                          doc="""Get/set the integrand""")
#
#
#     def to_dict(self, minimal=False):
#
#         data = super(Function2D, self).to_dict(minimal)
#
#         if not minimal:
#             data['extra_setup'] = {'integrand': self.integrand.path}
#
#         return data



#!/usr/bin/env python

import ctypes.util
import glob
import os
import re
import sys
import subprocess

from packaging import version as packaging_version
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext as _build_ext

import versioneer

# This is needed to use numpy in this module, and should work whether or not numpy is
# already installed. If it's not, it will trigger an installation

_default_xspec_version = "12.15.1"  # default when installing xspec according following
# https://heasarc.gsfc.nasa.gov/docs/software/conda.html


class My_build_ext(_build_ext):

    def finalize_options(self):

        _build_ext.finalize_options(self)

        import numpy

        self.include_dirs.append(numpy.get_include())
        self.include_dirs.append("astromodels/xspec/include")
        conda_prefix = os.environ.get("CONDA_PREFIX")
        if conda_prefix is not None:
            conda_include_path = os.path.join(conda_prefix, "include")
            self.include_dirs.append(conda_include_path)


def sanitize_lib_name(library_path):
    """Get a fully-qualified library name, like /usr/lib/libgfortran.so.3.0,
    and returns the lib name needed to be passed to the linker in the -l option
    (for example gfortran)

    :param library_path:
    :return:
    """

    lib_name = os.path.basename(library_path)

    # Some regexp magic needed to extract in a system-independent (mac/linux)
    # way the library name

    tokens = re.findall(r"lib(.+)(\.so|\.dylib|\.a|\.la)(.+)?", lib_name)

    if not tokens:
        msg = f"Attempting to find {lib_name} in directory {library_path}"
        msg += " but there are no libraries in this directory"

        raise RuntimeError(msg)

    return tokens[0][0]


def find_library(library_root, additional_places=None):
    """Returns the name of the library without extension.

    :param library_root: root of the library to search, for example
        "cfitsio_" will match libcfitsio_1.2.3.4.so
    :return: a tuple of (library_name, library_dir, full_path) where:
        - library_name: the name to be passed to the linker in the
          -l option
        - library_dir: the directory path (None if in system paths)
        - full_path: the full path to the library file (used when
          unversioned symlink is missing)
    """

    # find_library searches for all system paths in a system independent way (but NOT
    # those defined in LD_LIBRARY_PATH or DYLD_LIBRARY_PATH)

    first_guess = ctypes.util.find_library(library_root)

    if first_guess is not None and library_root == "gfortran":

        # Found in one of the system paths

        if sys.platform.lower().find("linux") >= 0:

            # On linux the linker already knows about these paths, so we
            # can return None as path

            return sanitize_lib_name(first_guess), None, None

        elif sys.platform.lower().find("darwin") >= 0:

            # On Mac we still need to return the path, because the linker sometimes
            # does not look into it

            return sanitize_lib_name(first_guess), os.path.dirname(first_guess), None

        else:

            # Windows is not supported

            raise NotImplementedError("Platform %s is not supported" % sys.platform)

    else:

        # could not find it. Let's examine LD_LIBRARY_PATH or DYLD_LIBRARY_PATH
        # (if they sanitize_lib_name(first_guess), are not defined, possible_locations
        # will become [""] which will be handled by the next loop)

        if sys.platform.lower().find("linux") >= 0:

            # Unix / linux

            possible_locations = os.environ.get("LD_LIBRARY_PATH", "").split(":")

        elif sys.platform.lower().find("darwin") >= 0:

            # Mac

            possible_locations = os.environ.get("DYLD_LIBRARY_PATH", "").split(":")

        else:

            raise NotImplementedError("Platform %s is not supported" % sys.platform)

        if additional_places is not None:

            possible_locations.extend(additional_places)

        # Now look into the search paths

        library_name = None
        library_dir = None
        library_full_path = None

        for search_path in possible_locations:

            if search_path == "":
                # This can happen if there are more than one :, or if nor
                # LD_LIBRARY_PATH nor DYLD_LIBRARY_PATH are defined (because of
                # the default use above for os.environ.get)

                continue

            results = glob.glob(os.path.join(search_path, f"lib{library_root}*"))

            if len(results) >= 1:

                # Results contain things like libXS.so, libXSPlot.so, libXSpippo.so
                # If we are looking for libXS.so, we need to make sure that we get the
                # right one!

                for result in results:

                    if (
                        re.match(
                            f"lib{library_root}" + r"[\-_\.]([0-9])*\d*(\.[0-9]\d*)*",
                            os.path.basename(result),
                        )
                        is None
                    ):

                        continue

                    else:

                        # FOUND IT

                        # This is the full path of the library, like
                        # /usr/lib/libcfitsio_1.2.3.4

                        library_full_path = result
                        library_name = result
                        library_dir = search_path

                        break

            else:

                continue

            if library_name is not None:
                break

        if library_name is None:

            return None, None, None

        else:

            # Sanitize the library name to get from the fully-qualified path
            # to just the library name (/usr/lib/libgfortran.so.3.0 becomes
            # gfortran)

            sanitized_name = sanitize_lib_name(library_name)

            # Check if the unversioned symlink exists
            # If not, we'll need to pass the full path to the linker
            if sys.platform.lower().find("linux") >= 0:
                extension = ".so"
            elif sys.platform.lower().find("darwin") >= 0:
                extension = ".dylib"
            else:
                extension = ".so"

            unversioned_lib = os.path.join(
                library_dir, f"lib{sanitized_name}{extension}"
            )

            if os.path.exists(unversioned_lib):
                # Unversioned symlink exists, use normal linking
                return sanitized_name, library_dir, None
            else:
                # No unversioned symlink, return the full path for direct
                # linking
                return sanitized_name, library_dir, library_full_path


def get_xspec_conda_version():
    """Get the version string from conda"""
    try:
        lines = (
            subprocess.check_output(["conda", "list", "-f", "xspec"])
            .decode()
            .split("\n")
        )
    except subprocess.CalledProcessError:
        lines = subprocess.check_output(["conda", "list", "-f", "xspec"]).split("\n")
    for l in lines:
        if not l:
            continue
        if l[0] == "#":
            continue
        tokens = l.split()
        return tokens[1]
    return None


def setup_xspec():

    skip_xspec = os.environ.get("SKIP_XSPEC")
    headas_root = os.environ.get("HEADAS")
    conda_prefix = os.environ.get("CONDA_PREFIX")

    if skip_xspec is not None:

        print(
            "The SKIP_XSPEC env variable was set. Xspec support will not be installed."
        )
        return None

    if headas_root is None:

        # See, maybe we are running in Conda

        if conda_prefix is None:

            # Maybe this is a Conda build

            conda_prefix = os.environ.get("PREFIX")

        if conda_prefix is not None:

            # Yes, this is Conda
            # Let's see if the package xspec-modelsonly has been installed by checking
            # whether one of the Xspec libraries exists within conda
            conda_lib_path = os.path.join(conda_prefix, "lib")
            this_lib, this_lib_path, full_lib_library = find_library(
                "XSFunctions", additional_places=[conda_lib_path]
            )

            if this_lib is None:

                # No, there is no library in Conda
                msg = "No xspec-modelsonly package has been installed in Conda. Xspec"
                msg += " support will not be installed"
                print(msg)

                print("Was looking into %s" % conda_lib_path)

                return None

            else:
                msg = (
                    "WARN: The xspec-modelsonly package has been installed"
                    " in Conda, but it's no longer supported."
                    " Xspec support will not be installed"
                )
                print(msg)

                return None

                # Set up the HEADAS variable so that the following will find the
                # libraries
                # headas_root = conda_prefix

        else:

            print("No HEADAS env. variable set. Xspec support will not be installed ")

            return None

    print("HEADAS env. variable detected. Will compile the Xspec extension.")
    print(
        "NOTICE: If you have issues, manually set the environment variable "
        "XSPEC_INC_PATH to the location of the XSPEC headers"
    )
    msg = "If you are still having issues, unset HEADAS before installing and"
    msg += "contact the support team"
    print(msg)

    xspec_version = get_xspec_conda_version()

    if xspec_version is not None:

        print("Found XSPEC version %s in Conda" % xspec_version)

    else:

        print("No XSPEC installation found in Conda")
        print("Xspec was likely compiled from source.")

        xspec_version = os.environ.get("ASTRO_XSPEC_VERSION")

        if xspec_version is None:
            print("WARN: You have not specified an XSPEC version with the ")
            print("WARN: environment variable ASTRO_XSPEC_VERSION")
            print(f"WARN: we will assume you have {_default_xspec_version}")
            print(
                "If you are using a different version of XSPEC, please set"
                " the environment variable ASTRO_XSPEC_VERSION to the "
                "version of XSPEC you are using"
            )

            xspec_version = _default_xspec_version

        else:

            print(f"You have specified the XSPEC version {xspec_version}")

    xspec_version = packaging_version.Version(xspec_version)

    if xspec_version < packaging_version.Version("12.12.0"):
        msg = "WARN: XSPEC version is less than 12.12.0, which is the minimal"
        msg += " supported version for astromodels"
        print(msg)
        return None
    elif xspec_version > packaging_version.Version("12.15.1"):
        msg = "WARN: XSPEC version is greater than 12.15.1, which is the"
        msg += " maximal supported version for astromodels"
        print(msg)
        return None

    macros = []
    # I am not sure what the naming of the XSPEC components are,
    # but let's stick with major, minor, and patch.
    for major, minor, patch in [
        (12, 12, 0),
        (12, 12, 1),
        (12, 13, 0),
        (12, 13, 1),
        (12, 14, 0),
        (12, 14, 1),
        (12, 15, 0),
        (12, 15, 1),
    ]:

        version = "{}.{}.{}".format(major, minor, patch)

        macro = "XSPEC_{}_{}_{}".format(major, minor, patch)

        if xspec_version >= packaging_version.Version(version):
            macros += [(macro, None)]

    print(macros)

    # Make sure these libraries exist and are linkable right now
    # (they need to be in LD_LIBRARY_PATH or DYLD_LIBRARY_PATH or in one of the system
    # paths)

    libraries_root = [
        "XSFunctions",
        "XSModel",
        "XSUtil",
        "XS",
        "cfitsio",
        "CCfits",
        "wcs",
        "gfortran",
    ]

    libraries = []
    library_dirs = []
    extra_objects = []  # For libraries without unversioned symlinks

    for lib_root in libraries_root:

        this_library, this_library_path, full_lib_path = find_library(
            lib_root, additional_places=[os.path.join(headas_root, "lib")]
        )

        if this_library is None:

            raise IOError(
                "Could not find library %s. Impossible to compile Xspec" % lib_root
            )

        else:

            print("Found library %s in %s" % (this_library, this_library_path))

            if full_lib_path is not None:
                # No unversioned symlink exists, pass the full path directly
                print(
                    "Warning: No unversioned symlink found for %s, "
                    "using full path %s" % (lib_root, full_lib_path)
                )
                extra_objects.append(full_lib_path)
            else:
                # Normal case: unversioned symlink exists, use -l linking
                libraries.append(this_library)

            if this_library_path is not None:
                # This library is not in one of the system path library, we need to add
                # it to the -L flag during linking. Let's put it in the library_dirs
                # list which will be used in the Extension class

                library_dirs.append(this_library_path)

    # try to manually add on the include directory

    header_paths = []

    if library_dirs:

        # grab it from the lib assuming that it is one up
        xspec_path, _ = os.path.split(library_dirs[0])
        include_path = os.path.join(xspec_path, "include")

        header_paths.append(include_path)

    # let's be sure to add the conda include directory

    if conda_prefix is not None:

        conda_include_path = os.path.join(conda_prefix, "include")
        header_paths.append(conda_include_path)

    # check if there are user set the location of the xspec headers:

    xspec_headers_path = os.environ.get("XSPEC_INC_PATH")

    if xspec_headers_path is not None:

        print("You have set XSPEC_INC_PATH=%s" % xspec_headers_path)

        header_paths.append(xspec_headers_path)

    # Remove duplicates from library_dirs and header_paths

    library_dirs = list(set(library_dirs))
    header_paths = list(set(header_paths))

    print("header paths:")
    for h in header_paths:

        print(f"{h}")

    # Configure the variables to build the external module with the C/C++ wrapper

    ext_modules_configuration = [
        Extension(
            "astromodels.xspec._xspec",
            [
                "astromodels/xspec/src/_xspec.cc",
            ],
            include_dirs=header_paths,
            libraries=libraries,
            library_dirs=library_dirs,
            runtime_library_dirs=library_dirs,
            extra_compile_args=[],
            # Add full paths for libraries without unversioned symlinks
            extra_objects=extra_objects,
            define_macros=macros,
        ),
    ]

    return ext_modules_configuration


# Normal packages

packages = [
    "astromodels",
    "astromodels/core",
    "astromodels/functions",
    "astromodels/functions/functions_1D",
    "astromodels/functions/dark_matter",
    "astromodels/sources",
    "astromodels/utils",
    "astromodels/xspec",
    "astromodels/tests",
]

# Check whether we can compile Xspec support
ext_modules_configuration = setup_xspec()

# Add the node_ctype module


setup(
    cmdclass=versioneer.get_cmdclass({"build_ext": My_build_ext}),
    version=versioneer.get_version(),
    ext_modules=ext_modules_configuration,
)



# This file contains some defaults, like locations of files, which should not
# change much but benefits anyway of being in one central location

import os
from importlib.resources import files
from pathlib import Path
from typing import Optional, Union

import numpy as np

_custom_config_path = os.environ.get("ASTROMODELS_CONFIG")

copy_if_needed: Optional[bool]

if np.lib.NumpyVersion(np.__version__) >= "2.0.0":
    copy_if_needed = None
else:
    copy_if_needed = False


def _get_data_file_path(data_file: Union[str, Path]) -> Path:
    """Returns the absolute path to the required data files.

    :param data_file: relative path to the data file (str or Path), relative to
    astromodels/data/.
    Example: "dark_matter/gammamc_dif.dat" or Path("dark_matter/gammamc_dif.dat")
    :return: absolute path of the data file as a Path object
    """
    data_file = Path(data_file)

    try:
        resource_path = files("astromodels").joinpath("data", *data_file.parts)

        if not resource_path.is_file():
            raise FileNotFoundError

    except Exception:
        raise IOError(
            f"Could not read or find data file {data_file}. "
            "Try reinstalling astromodels. "
            f"If this does not fix your problem, open an issue on github."
        )

    else:
        return Path(resource_path).resolve()


def get_user_path() -> Path:

    user_path: Path = Path.home() / ".astromodels"

    if not user_path.exists():

        user_path.mkdir(parents=True)

    return user_path


def get_user_data_path() -> Path:

    user_data: Path = get_user_path() / "data"

    # Create it if doesn't exist
    if not user_data.exists():

        user_data.mkdir(parents=True)

    return user_data


def get_path_of_user_config() -> Path:

    if _custom_config_path is not None:

        config_path: Path = Path(_custom_config_path)

    config_path: Path = Path().home() / ".config" / "astromodels"

    if not config_path.exists():

        config_path.mkdir(parents=True)

    return config_path
